# -*- coding: utf-8 -*-
#
# Zimbra Corporation Server 4 UCS
#
# Copyright 2016-2017 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import os
import json
import logging
import urllib2

import ldap

from pythonzimbra.tools.auth import authenticate
from pythonzimbra.communication import Communication
#from pythonzimbra.request_json import RequestJson
#from pythonzimbra.response_json import ResponseJson
from pythonzimbra.exceptions.auth import AuthenticationFailed

import univention.admin.uldap
from univention.lib.i18n import Translation
from univention.config_registry import ConfigRegistry

from zimbra4ucs.errors import ZimbraAuthenticationFailed, ZimbraResponseError, ZimbraObjectAlreadyExists, ZimbraNotInitialized, ZimbraConnectionFailed, ZimbraNoSuchObject

_ = Translation('univention-zimbra').translate
ucr = ConfigRegistry()
ucr.load()
logger = logging.getLogger('zimbra4ucs')


class ZimbraConfiguration(object):

	ZIMBRA_URL = "https://%s:7071/service/admin/soap"
	CONFIGURATION_FILE = '/etc/zimbra4ucs/configuration.json'

	def __init__(self, hostname, username, password):
		self.hostname = hostname
		self.username = username
		self.password = password
		self.connection = None
		self.url = self.ZIMBRA_URL % (self.hostname,)

	def authenticate(self):
		self.connection = ZimbraConnection(self.url, self.username, self.password)
		self.connection.authenticate()

	def configure(self, maildomain):
		self.authenticate()
		self.create_default_ucs_account(maildomain)
		self.configure_mail_domains(maildomain)
		# TODO: create a UCS mail/domain for every existing mail domain?

	def create_default_ucs_account(self, maildomain):
		from zimbra4ucs.models import ZimbraAccount
		username = 'zimbra4ucs@%s' % (maildomain,)
		with open('/etc/zimbra4ucs.secret') as fd:
			password = fd.read().strip()
		dn = 'uid=zimbra4ucs,cn=users,%s' % (ucr['ldap/base'],)
		attrs = {}
		try:
			lo, po = univention.admin.uldap.getMachineConnection()
			attrs = lo.get(dn)
		except ldap.LDAPError as exc:
			logger.error('Could not query LDAP information for initial account. Fallback to some default values. Error: %s', exc)
		values = dict(
			ZimbraAccount.map_ldap_attributes(attrs),
			zimbraIsSystemResource=True,
			zimbraIsSystemAccount=True,
			zimbraIsAdminAccount=True,
			zimbraAuthLdapExternalDn=dn,
			zimbraAccountStatus='active',
			password=password,
			name=username,
		)
		account = ZimbraAccount(self.connection, **values)
		try:
			zid = account.get_id(username)
			account.password = None
			account.modify(dict(values, password=password))
		except ZimbraNoSuchObject as exc:
			logger.info('User does not exist, going to create ...')
			response = account.create()
			zid = response['id']
			username = response['name']
		if zid and username:
			self.store_configuration(dict(
				hostname=self.hostname,
				username=username,
				password=password,
				uuid=zid,
			))

	def configure_mail_domains(self, maildomain):
		# TODO: do this for every existing domain?
		from zimbra4ucs.models import ZimbraMailDomains
		values = dict((key, None) for key in ZimbraMailDomains.get_authentication_defaults().keys())
		domain = ZimbraMailDomains(self.connection, name=maildomain, **values)
		domain.modify(dict(domain.values(), **ZimbraMailDomains.get_authentication_defaults()))

	@classmethod
	def store_configuration(cls, data):
		fd = os.open(cls.CONFIGURATION_FILE, os.O_WRONLY | os.O_CREAT, 0600)
		os.write(fd, json.dumps(data))
		os.close(fd)

	@classmethod
	def is_initialized(cls):
		return bool(cls.get_configuration())

	@classmethod
	def get_configuration(cls):
		try:
			with open(cls.CONFIGURATION_FILE) as fd:
				return json.load(fd)
		except (ValueError, EnvironmentError):
			return {}

	@classmethod
	def get_connection(cls):
		config = cls.get_configuration()
		try:
			self = cls(config['hostname'], config['username'], config['password'])
		except KeyError:
			raise ZimbraNotInitialized()
		self.authenticate()
		return self.connection


class ZimbraConnection(object):

	def __init__(self, url, username, password, timeout=None):
		self.url = url
		self.username = username
		self.password = password
		self.timeout = timeout
		self.auth_token = None

	def authenticate(self):
		try:
			self.auth_token = authenticate(
				self.url,
				self.username,
				self.password,
				admin_auth=True,
				request_type="json",
				timeout=self.timeout,
				raise_on_error=True,
			)
		except AuthenticationFailed as exc:
			logger.error('Error during authentication: %s', exc)
			raise ZimbraAuthenticationFailed(_('Could not authenticate.'))
		except urllib2.URLError as exc:
			logger.error('Connection Error during authentication: %s', exc)
			raise ZimbraConnectionFailed(_('Could not authenticate at zimbra server at %s: %s') % (self.url, exc,))

	def request(self, method, data):
		communication = Communication(self.url, self.timeout)
		request = communication.gen_request(token=self.auth_token)
		request.add_request(method, data, 'urn:zimbraAdmin')

		logger.debug('request: %s' % request.get_request())

		try:
			response = communication.send_request(request)
			#logger.debug('response: %s' % response.get_response())
		except urllib2.URLError as exc:
			logger.error('Connection Error: %s', exc)
			raise ZimbraConnectionFailed(_('Could not connect to zimbra server at %s: %s') % (self.url, exc,))

		if response.is_fault():
			code = response.get_fault_code()
			cls = {
				'account.ACCOUNT_EXISTS': ZimbraObjectAlreadyExists,
				'account.DOMAIN_EXISTS': ZimbraObjectAlreadyExists,
				'account.NO_SUCH_ACCOUNT': ZimbraNoSuchObject,
				'service.INVALID_REQUEST': ZimbraResponseError,
			}.get(code, ZimbraResponseError)
			raise cls(response.get_fault_message(), code)
		return response
