import logging

from univention.config_registry import ConfigRegistry

from pythonzimbra.tools.dict import zimbra_to_python

from zimbra4ucs.errors import ZimbraResponseEmpty, ZimbraObjectAlreadyExists, ZimbraNoSuchObject
from zimbra4ucs import ZimbraConfiguration

logger = logging.getLogger('zimbra4ucs')
ucr = ConfigRegistry()
ucr.load()


class Attribute(object):

	def __init__(self, name, ldap_attribute=None, multivalue=False, required=False, identifies=False, default=None, map_function=None, unmap_function=None, encode_function=None, decode_function=None):
		self.name = name
		self.ldap_attribute = ldap_attribute
		self.multivalue = multivalue
		self.default = default
		self.identifies = identifies
		self.required = required
		self.map_function = map_function
		self.unmap_function = unmap_function
		self.encode_function = encode_function
		self.decode_function = decode_function


def encode_boolean(data):
	return 'TRUE' if data else 'FALSE'


def decode_boolean(data):
	return {'TRUE': True, '1': True, 'FALSE': False, '0': False, }.get(data)


def encode_account_status(data):
	if data in ["active", "locked", "closed", "pending"]:
		return data


class ZimbraObject(object):

	connection = None
	attributes = []
	result_key = None

	def __init__(self, connection=None, **attributes):
		self.connection = connection
		for attribute in self.attributes:
			value = attributes.pop(attribute.name, attribute.default)
			setattr(self, attribute.name, value)
		if attributes:
			raise KeyError(', '.join(attributes.keys()))

	def get(self, *attrs, **by):
		"""search for object by attribute. queries the attributes"""
		key, val = by.items()[0]
		data = {'attrs': ','.join(attrs), self.result_key: {'by': key, '_content': val}}
		return self.request('%sRequest' % self.method_get, data)['%sResponse' % self.method_get][self.result_key]

	def get_id(self, name):
		"""get the real UUID / zimbraId of the object by its name"""
		return self.get('zimbraId', name=name)['id']

	def create(self):
		"""create the object with all current set values"""
		data = self.request('%sRequest' % self.method_create, self.dict())['%sResponse' % self.method_create]
		if self.result_key:
			data = data[self.result_key]
		return data

	def modify(self, values):
		"""modify the object by building a diff between the current values and the given new values"""
		data = self.diff(values)
		data['id'] = self.get_id(self.name)
		return self.request('%sRequest' % self.method_modify, data)['%sResponse' % self.method_modify][self.result_key]

	def delete(self):
		"""delete the object"""
		data = dict(id=self.get_id(self.name))
		return self.request('%sRequest' % self.method_delete, data)['%sResponse' % self.method_delete]

	def rename(self, values):
		"""rename the object by building a diff of the current identifying name attribute"""
		data = dict(id=self.get_id(self.name), newName=self.dict()['name'])
		return self.request('%sRequest' % self.method_rename, data)['%sResponse' % self.method_rename][self.result_key]

	def get_all(self):
		"""get all objects of this type"""
		return self.request('%sRequest' % self.method_query, {})['%sResponse' % self.method_query][self.result_key]

	def request(self, method, data=None):
		logger.info('Request %r with %r', method, data)
		return self.connection.request(method, data).get_response()

	@classmethod
	def from_dict(cls, data):
		return cls(**zimbra_to_python(data))

	def dict(self):
		attrs = self.get_attributes()
		return self.to_dict(attrs)

	def to_dict(self, attrs):
		data = dict((key, value) for attr, key, value in attrs if attr.identifies)
		data['a'] = [dict(n=key, _content=value) for attr, key, value in attrs if not attr.identifies]
		return data

	def diff(self, values):
		current = dict((x[1], x[2]) for x in self.get_attributes())
		new = dict((x[1], x[2]) for x in type(self)(self.connection, **values).get_attributes())
		diff = []
		for attr in self.attributes:
			if new.get(attr.name) != current.get(attr.name):
				value = new.get(attr.name)
				if attr.encode_function:
					value = attr.encode_function(value)
				diff.append((attr, attr.name, value))
		return self.to_dict(diff)

	def get_attributes(self):
		attributes = []
		for attr in self.attributes:
			value = getattr(self, attr.name, attr.default)
			if attr.encode_function:
				value = attr.encode_function(value)
			if value is not None:
				attributes.append((attr, attr.name, value))
		return attributes

	def values(self):
		return dict((attr.name, getattr(self, attr.name)) for attr in self.attributes)

	@classmethod
	def ldap_create(cls, dn, new):
		new['dn'] = dn
		try:
			data = cls(cls.get_connection(), **cls.map_ldap_attributes(new)).create()
		except ZimbraObjectAlreadyExists as exc:
			logger.error('Object already exists: %s', exc)
			return
		return data['id']  # TODO: store as attribute

	@classmethod
	def ldap_rename(cls, dn, old_dn, old, new):
		new['dn'] = dn
		try:
			return cls(cls.get_connection(), **cls.map_ldap_attributes(old)).rename(cls.map_ldap_attributes(new))
		except ZimbraNoSuchObject as exc:
			logger.error('Could not rename object. Object does not exists: %s', exc)

	@classmethod
	def ldap_modify(cls, dn, new, old):
		new['dn'] = dn
		try:
			return cls(cls.get_connection(), **cls.map_ldap_attributes(old)).modify(cls.map_ldap_attributes(new))
		except ZimbraNoSuchObject as exc:
			logger.error('Could not modify object. Object does not exists: %s', exc)

	@classmethod
	def ldap_delete(cls, dn, old):
		old['dn'] = dn
		try:
			return cls(cls.get_connection(), **cls.map_ldap_attributes(old)).delete()
		except ZimbraNoSuchObject as exc:
			logger.warn('Could not remove object. Object does not exists: %s', exc)

	@classmethod
	def get_connection(cls):
		if cls.connection is None:
			cls.connection = ZimbraConfiguration.get_connection()
		return cls.connection

	@classmethod
	def map_ldap_attributes(cls, attrs):
		props = {}
		for prop in cls.attributes:
			value = attrs.get(prop.ldap_attribute)
			if value and not prop.multivalue:
				value = value[0]
			if prop.decode_function:
				value = prop.decode_function(value)
			props[prop.name] = value
		return props


class ZimbraMailDomains(ZimbraObject):

	result_key = 'domain'
	method_create = 'CreateDomain'
	method_modify = 'ModifyDomain'
	method_rename = 'RenameDomain'
	method_delete = 'DeleteDomain'
	method_query = 'GetAllDomains'
	method_get = 'GetDomain'

	attributes = [
		Attribute('name', 'cn', identifies=True, required=True),
		Attribute('description', 'description'),
		Attribute('zimbraNotes'),

		Attribute('zimbraAuthMech', default='ldap', required=True),  # ['ad', 'ldap']
		Attribute('zimbraAuthLdapURL', required=True),
		Attribute('zimbraAuthLdapBindDn', required=True),
		Attribute('zimbraAuthLdapSearchBase', default=''),
		Attribute('zimbraAuthLdapSearchBindDn'),
		Attribute('zimbraAuthLdapSearchBindPassword'),
		Attribute('zimbraAuthLdapSearchFilter'),
		Attribute('zimbraAuthLdapStartTlsEnabled', encode_function=encode_boolean, decode_function=decode_boolean),
	]

	def get_all(self):
		try:
			domains = super(ZimbraMailDomains, self).get_all()
		except KeyError:
			raise ZimbraResponseEmpty()

		if not isinstance(domains, (list, tuple)):
			domains = [domains]
		return domains

	def create(self):
		for key, value in self.get_authentication_defaults().iteritems():
			setattr(self, key, value)
		return super(ZimbraMailDomains, self).create()

	def rename(self, values):
		self.delete()
		type(self)(self.connection, **values).create()

	def modify(self, values):
		values = dict(values, **self.get_authentication_defaults())
		super(ZimbraMailDomains, self).modify(values)

	@classmethod
	def get_authentication_defaults(cls):
		dn = 'uid=zimbra4ucs,cn=users,%s' % (ucr['ldap/base'],)
		with open('/etc/zimbra4ucs.secret') as fd:
			password = fd.read().strip()
		return dict(
			zimbraAuthMech='ldap',
			zimbraAuthLdapURL='ldap://%s:%s' % (ucr.get('ldap/master'), 7389),
			zimbraAuthLdapBindDn=None,  # we use the search filter instead
			zimbraAuthLdapSearchBase=ucr.get('ldap/base'),
			zimbraAuthLdapSearchBindDn=dn,
			zimbraAuthLdapSearchBindPassword=password,
			zimbraAuthLdapSearchFilter='(&(mailPrimaryAddress=%n)(objectClass=person))',
			zimbraAuthLdapStartTlsEnabled=True,
		)


class ZimbraAccount(ZimbraObject):

	result_key = 'account'
	method_create = 'CreateAccount'
	method_modify = 'ModifyAccount'
	method_rename = 'RenameAccount'
	method_delete = 'DeleteAccount'
	method_query = 'GetAllAcounts'
	method_get = 'GetAccount'

	attributes = [
		Attribute('name', 'mailPrimaryAddress', required=True, identifies=True),
		Attribute('password', identifies=True),
		Attribute('gn', 'givenName'),
		Attribute('sn', 'sn', required=True),
		Attribute('displayName', 'displayName'),
		Attribute('description', 'description'),
		Attribute('telephoneNumber', 'telephoneNumber'),
		Attribute('homePhone'),
		Attribute('mobile'),
		Attribute('pager'),
		Attribute('facsimileTelephoneNumber'),
		Attribute('company', 'o'),
		Attribute('title', 'title'),
		Attribute('street', 'street'),  # address
		Attribute('l', 'l'),  # location
		Attribute('st', 'st'),  # state
		Attribute('postalCode', 'postalCode'),
		Attribute('co'),  # country
		Attribute('zimbraCOSId', 'zimbraCOSId'),  # $uuid
		Attribute(
			'zimbraAuthLdapExternalDn',
			# 'dn',  # don't sync this attribute currently, as we don't handle ldap-rename and ldap-move.
			required=True,
		),
		Attribute('zimbraHideInGal', 'zimbraHideInGal', default=False, encode_function=encode_boolean, decode_function=decode_boolean),
		Attribute('zimbraAccountStatus', 'zimbraAccountStatus', default="active", encode_function=encode_account_status),  # ["active", "locked", "closed", "pending"]
		Attribute('zimbraIsSystemAccount', 'zimbraIsSystemAccount', required=True, default=False, encode_function=encode_boolean, decode_function=decode_boolean),
		Attribute('zimbraIsAdminAccount', 'zimbraIsAdminAccount', required=True, default=False, encode_function=encode_boolean, decode_function=decode_boolean),
		Attribute('zimbraIsSystemResource', 'zimbraIsSystemResource', required=True, default=False, encode_function=encode_boolean, decode_function=decode_boolean),
	]

	def create(self):
		return super(ZimbraAccount, self).create()

	def modify(self, values):
		#if not self.encode_account_status(self.zimbraAccountStatus):
		#	return self.delete()
		return super(ZimbraAccount, self).modify(values)


class ZimbraCheckAuthConfig(ZimbraObject):

	attributes = [
		Attribute('name', identifies=True),
		Attribute('password', identifies=True),
		Attribute('zimbraAuthMech', default='ldap', required=True),  # ['ad', 'ldap']
		Attribute('zimbraAuthLdapURL', required=True),
		Attribute('zimbraAuthLdapBindDn', required=True),
		Attribute('zimbraAuthLdapSearchBase', default=''),
		Attribute('zimbraAuthLdapSearchBindDn'),
		Attribute('zimbraAuthLdapSearchBindPassword'),
		Attribute('zimbraAuthLdapSearchFilter'),
		Attribute('zimbraAuthLdapStartTlsEnabled', encode_function=encode_boolean, decode_function=decode_boolean),
	]

	method_create = 'CheckAuthConfig'
