# -*- coding: utf-8 -*-
#
# UCS import
#
# Copyright 2017 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import os
import logging
import cPickle

from zimbra4ucs.models import ZimbraAccount, ZimbraMailDomains

logger = logging.getLogger('zimbra4ucs')


class UcsImport(object):

	connection = None

	@classmethod
	def get_changes(cls, new, old):
		changes = dict()
		for i in new:
			if i not in old:
				changes[i] = (new[i], None)
			else:
				if set(new[i]) != set(old[i]):
					changes[i] = (new[i], old[i])
		for i in old:
			if i not in new:
				changes[i] = (None, old[i])
		return changes

	@classmethod
	def process_files(cls, directory, dry_run):
		cls.dry_run = dry_run
		files = os.listdir(directory)
		files.sort()
		for fil in files:
			fil = os.path.join(directory, fil)
			if fil.endswith('tmp') or os.path.isdir(fil) or not os.path.isfile(fil):
				continue
			# load data
			try:
				with open(fil) as f:
					p = cPickle.Unpickler(f)
					dn, new, old, old_dn = p.load()
			except Exception as ex:
				logger.exception('could not load listener pickle file %s: %r' % (fil, ex))
				continue
			logger.info('processing file: %s' % fil)
			logger.debug('dn: %s' % dn)
			logger.debug('old_dn: %s' % old_dn)
			logger.debug('new: %s' % new)
			logger.debug('old: %s' % old)
			if cls.is_create(new, old, dn, old_dn):
				logger.info('create %s' % dn)
				if not dry_run:
					cls.create(dn, new)
			elif cls.is_rename(new, old, dn, old_dn):
				logger.info('rename %s to %s' % (old_dn, dn))
				changes = cls.get_changes(new, old)
				logger.info('changes %s' % changes)
				if not dry_run:
					cls.rename(dn, old_dn, old, new)
			elif cls.is_modify(new, old, dn, old_dn):
				logger.info('modify %s' % dn)
				changes = cls.get_changes(new, old)
				logger.info('changes %s' % changes)
				if not dry_run:
					cls.modify(dn, new, old)
			elif cls.is_remove(new, old, dn, old_dn):
				logger.info('delete %s' % dn)
				if not dry_run:
					cls.delete(dn, old)
			else:
				logger.info('ignore change for %s' % dn)

			if not dry_run:
				os.remove(fil)


class ImportUser(UcsImport):

	@classmethod
	def is_active(cls, obj):
		logger.debug('status %s' % obj.get('zimbraAccountStatus'))
		logger.debug('mail %s' % obj.get('mailPrimaryAddress'))
		if obj.get('zimbraAccountStatus', ['none'])[0].lower() != 'none':
			if obj.get('mailPrimaryAddress', None):
				return True
		return False

	@classmethod
	def is_create(cls, new, old, dn, old_dn):
		if new and not old:
			if cls.is_active(new):
				return True
		if new and old:
			if cls.is_active(new) and not cls.is_active(old):
					return True
		return False

	@classmethod
	def is_rename(cls, new, old, dn, old_dn):
		# rename (mailPrimaryAddress has been changed) is checked in backend
		return False

	@classmethod
	def is_modify(cls, new, old, dn, old_dn):
		if new and old:
			if cls.is_active(new) and cls.is_active(old):
				for i in ZimbraAccount.attributes:
					if i.name in new:
						return True
		return False

	@classmethod
	def is_remove(cls, new, old, dn, old_dn):
		if old and new:
			if not cls.is_active(new) and cls.is_active(old):
				return True
		if old and not new:
			if cls.is_active(old):
				return True
		return False

	model = ZimbraAccount
	create = model.ldap_create
	rename = model.ldap_rename
	modify = model.ldap_modify
	delete = model.ldap_delete


class ImportDomain(UcsImport):

	@classmethod
	def is_create(cls, new, old, dn, old_dn):
		if new and not old:
			return True
		return False

	@classmethod
	def is_rename(cls, new, old, dn, old_dn):
		if old and new:
			if old.get('cn')[0] != new.get('cn')[0]:
				return True
		return False

	@classmethod
	def is_modify(cls, new, old, dn, old_dn):
		if old and new:
			if old.get('cn')[0] == new.get('cn')[0]:
				return True
		return False

	@classmethod
	def is_remove(cls, new, old, dn, old_dn):
		if old and not new:
			return True
		return False

	model = ZimbraMailDomains
	create = model.ldap_create
	rename = model.ldap_rename
	modify = model.ldap_modify
	delete = model.ldap_delete
